"""Types for library users."""

import datetime as dt
import enum
from dataclasses import dataclass
from functools import cached_property
from typing import Literal


class TrafficLightColor(enum.StrEnum):
    """Color of a traffic light."""

    RED = enum.auto()
    YELLOW = enum.auto()
    GREEN = enum.auto()


UnknownT = Literal["UNKNOWN"]
UNKNOWN: UnknownT = "UNKNOWN"


@dataclass(frozen=True)
class Area:
    """An area (or sometimes residence, e.g. an apartment or parking space)."""

    name: str
    id: str
    data_available_since: dt.date
    data_available_until: dt.date


@dataclass(frozen=True)
class Quantity:
    """A representation of a physical quantity, by value and unit."""

    value: float
    unit: str


@dataclass(frozen=True)
class TrafficLightStatus:
    """Data returned via the traffic light page."""

    color: TrafficLightColor | UnknownT
    """Color of the traffic light."""

    current_energy_price: Quantity | UnknownT


@dataclass(frozen=True)
class MeterStatus:
    """Information about a meter and its current state."""

    name: str
    area: str
    meter_id: str
    timestamp: dt.datetime
    reading: Quantity


class ConsumptionType(enum.StrEnum):
    """Type of thing to consume (i.e. something that would be measured by a utility meter)."""

    ELECTRICITY = enum.auto()
    WATER_HOT = enum.auto()
    WATER_COLD = enum.auto()
    HEAT = enum.auto()


@dataclass(frozen=True)
class Consumption:
    """Indicates the amount of consumption at a specific time."""

    start: dt.datetime
    period: dt.timedelta

    value: float
    unit: str


@dataclass(frozen=True)
class PhotovoltaicSummary:
    """Information about the quarter's solar generation and power consumption in a given period."""

    start: dt.datetime
    """Start of the measurement period."""

    period: dt.timedelta
    """Duration of the measurement period."""

    consumption: Quantity
    """Energy consumption during the current period."""

    generation: Quantity
    """Energy generation (through PV cells) during the current period."""

    @cached_property
    def calculated_feed_into_grid(self) -> Quantity:
        """The (calculated) amount of energy being fed into the electrical grid."""
        if self.own_consumption is None:
            # We know: own_consumption = consumption_from_pv / pv_generation
            # own_consumption can not be calculated if pv_generation is 0.
            # But if pv_generation is 0, the feed into grid is also 0.
            return Quantity(value=0.0, unit=self.generation.unit)

        feed_factor = 1.0 - (self.own_consumption.value / 100.0)
        return Quantity(value=self.generation.value * feed_factor, unit=self.generation.unit)

    @cached_property
    def calculated_supply_from_grid(self) -> Quantity:
        """The (calculated) amount of energy being pulled from the electrical grid."""
        # We know: self_sufficiency = pv_usage / (pv_usage + supply_from_grid)
        #     and:      consumption = pv_usage + supply_from_grid
        # =>       supply_from_grid = consumption * (1 - self_sufficiency)
        if self.self_sufficiency is None:
            # self_sufficiency is only None when the PV production is 0.
            # In that case, we pull all power from the grid.
            return self.consumption

        supply_factor = 1.0 - (self.self_sufficiency.value / 100.0)
        return Quantity(self.consumption.value * supply_factor, unit=self.consumption.unit)

    self_sufficiency: Quantity | None
    """Self-sufficiency (the share of self-generated energy in consumed energy), usually in %."""

    own_consumption: Quantity | None
    """
    Own consumption (the share of self-consumed energy in self-generated energy), usually in %.
    """



"""Tests for the oocone.Enocoo.get_individual_consumption() method."""

import datetime as dt
import json

import pytest

from oocone import Auth, Enocoo
from oocone._internal.scrape_consumption import _CONSUMPTION_CLASSES as CONSUMPTION_CLASSES
from oocone.model import ConsumptionType
from tests import TIMEZONE
from tests.conftest import MOCK_API_DATASET_NAMES, MockApiParams


def _get_consumption_sum(  # noqa: PLR0913
    consumption_type: ConsumptionType,
    *,
    date: dt.date,
    interval: str,
    compensate_off_by_one: bool,
    area_id: str,
    mock_api_params: MockApiParams,
) -> float:
    def enocoo_response(request_date: dt.date) -> list:
        response_path = (
            mock_api_params.response_data_path
            / "getMeterDataWithParam.{consumption}.{area_id}.{date}.{interval}.php".format(  # noqa: UP032
                consumption=CONSUMPTION_CLASSES[consumption_type],
                area_id=area_id,
                date=request_date.isoformat(),
                interval=interval,
            )
        )

        with response_path.open("rb") as response_file:
            response = response_file.read()

        return json.loads(response)

    if consumption_type == ConsumptionType.HEAT and interval == "Tag":
        # Heat readings are integrated, so we need to calculate differences
        raw_readings, hours = enocoo_response(date)[:2]
        if raw_readings:  # readings are actually available
            readings_current_day = [raw_readings[0], raw_readings[-1] - raw_readings[0]]
        else:  # no readings for the current day are available
            readings_current_day = []
    else:
        readings_current_day, hours = enocoo_response(date)[:2]

        # Remove any next-day data that we might see
        while len(hours) >= 1 and hours[-1] == 0:
            hours.pop()
            readings_current_day.pop()

    if compensate_off_by_one:
        readings_next_day = enocoo_response(date + dt.timedelta(days=1))[0]
        relevant_readings = readings_current_day[1:] + readings_next_day[:1]
    else:
        relevant_readings = readings_current_day

    return sum(relevant_readings)


@pytest.mark.parametrize(
    "date",
    [
        pytest.param(dt.date(2024, 1, 1), id="regular day"),
        pytest.param(dt.date(2023, 10, 29), id="summer time to winter time"),
        pytest.param(dt.date(2024, 3, 31), id="winter time to summer time"),
    ],
)
@pytest.mark.parametrize(
    "consumption_type",
    [
        ConsumptionType.ELECTRICITY,
        ConsumptionType.WATER_COLD,
        ConsumptionType.WATER_HOT,
        ConsumptionType.HEAT,
    ],
)
@pytest.mark.parametrize("dataset", MOCK_API_DATASET_NAMES)
@pytest.mark.parametrize(
    "compensate_quirks",
    [
        pytest.param(False, id="uncompensated"),
        pytest.param(True, id="compensated"),
    ],
)
async def test_daily(  # noqa: PLR0913
    *,
    date: dt.date,
    consumption_type: ConsumptionType,
    compensate_quirks: bool,
    mock_auth: Auth,
    mock_api_params: MockApiParams,
    dataset: str,
) -> None:
    """Check that enocoo.get_individual_consumption returns daily data in expected format."""
    mock_api_params.dataset_name = dataset
    enocoo = Enocoo(mock_auth, TIMEZONE)

    for area in await enocoo.get_areas():
        consumption = await enocoo.get_individual_consumption(
            consumption_type=consumption_type,
            area_id=area.id,
            during=date,
            interval="day",
            compensate_off_by_one=compensate_quirks,
        )

        for hour in {cons.start.hour for cons in consumption}:
            period_sum = sum(
                [cons.period for cons in consumption if cons.start.hour == hour],
                start=dt.timedelta(),
            )
            assert period_sum == dt.timedelta(hours=1), (
                f"periods for hour {hour} should add up to one hour"
            )

        expected_sum = _get_consumption_sum(
            consumption_type,
            date=date,
            interval="Tag",
            compensate_off_by_one=compensate_quirks,
            area_id=area.id,
            mock_api_params=mock_api_params,
        )
        actual_sum = sum(cons.value for cons in consumption)
        assert actual_sum == expected_sum


@pytest.mark.parametrize("dataset", MOCK_API_DATASET_NAMES)
@pytest.mark.parametrize(
    "consumption_type",
    [
        ConsumptionType.ELECTRICITY,
        ConsumptionType.WATER_COLD,
        ConsumptionType.WATER_HOT,
        ConsumptionType.HEAT,
    ],
)
async def test_monthly(
    *,
    consumption_type: ConsumptionType,
    mock_auth: Auth,
    mock_api_params: MockApiParams,
    dataset: str,
) -> None:
    """Check that enocoo.get_individual_consumption returns monthly data in expected format."""
    date = dt.date(2024, 1, 1)
    mock_api_params.dataset_name = dataset
    enocoo = Enocoo(mock_auth, TIMEZONE)

    for area in await enocoo.get_areas():
        consumption = await enocoo.get_individual_consumption(
            consumption_type=consumption_type,
            area_id=area.id,
            during=date,
            interval="month",
        )

        expected_sum = _get_consumption_sum(
            consumption_type,
            date=date,
            interval="Monat",
            compensate_off_by_one=False,
            area_id=area.id,
            mock_api_params=mock_api_params,
        )
        actual_sum = sum(cons.value for cons in consumption)
        assert actual_sum == expected_sum


@pytest.mark.parametrize(
    "consumption_type",
    [
        ConsumptionType.ELECTRICITY,
        ConsumptionType.WATER_COLD,
        ConsumptionType.WATER_HOT,
        ConsumptionType.HEAT,
    ],
)
@pytest.mark.parametrize("dataset", MOCK_API_DATASET_NAMES)
async def test_yearly(
    consumption_type: ConsumptionType,
    mock_auth: Auth,
    dataset: str,
    mock_api_params: MockApiParams,
) -> None:
    """Check that enocoo.get_individual_consumption returns yearly data in expected format."""
    mock_api_params.dataset_name = dataset
    enocoo = Enocoo(mock_auth, TIMEZONE)

    for area in await enocoo.get_areas():
        consumption = await enocoo.get_individual_consumption(
            consumption_type=consumption_type,
            area_id=area.id,
            during=dt.date(2024, 1, 1),
            interval="year",
        )

        for reading in consumption:
            match reading.start.month:
                case 1 | 3 | 5 | 7 | 8 | 10 | 12:
                    assert reading.period == dt.timedelta(days=31)
                case 2:
                    assert reading.period == dt.timedelta(days=29)
                case 4 | 6 | 9 | 11:
                    assert reading.period == dt.timedelta(days=30)



"""Tests for the oocone.Enocoo.get_area_ids() method."""

import datetime as dt

import pytest

from oocone import Auth, Enocoo
from oocone.model import Area
from tests import TIMEZONE
from tests.conftest import MockApiParams


@pytest.mark.parametrize(
    ("dataset", "expected_areas"),
    [
        (
            "one_area",
            [
                Area(
                    name="H12W34",
                    id="123",
                    data_available_since=dt.date(2021, 1, 1),
                    data_available_until=dt.date(2022, 2, 2),
                )
            ],
        ),
        (
            "two_areas",
            [
                Area(
                    name="H12W34",
                    id="123",
                    data_available_since=dt.date(2021, 1, 1),
                    data_available_until=dt.date(2021, 1, 1),
                ),
                Area(
                    name="SP567 TNr:890",
                    id="124",
                    data_available_since=dt.date(2021, 1, 1),
                    data_available_until=dt.date(2021, 1, 1),
                ),
            ],
        ),
    ],
)
async def test_happy_path(
    dataset: str, expected_areas: list[Area], mock_auth: Auth, mock_api_params: MockApiParams
) -> None:
    """Check that enocoo.get_areas returns the areas as indicated in ownConsumption.php."""
    mock_api_params.dataset_name = dataset
    enocoo = Enocoo(mock_auth, TIMEZONE)

    areas = await enocoo.get_areas()
    assert areas == expected_areas



"""Tests suite for `oocone`."""

from pathlib import Path
from zoneinfo import ZoneInfo

from aiohttp import web, web_app
from aiohttp.test_utils import TestClient

MockApiClient = TestClient[web.Request, web_app.Application]
TESTS_DIR: Path = Path(__file__).parent
TIMEZONE = ZoneInfo("Europe/Berlin")

__all__ = [
    "TESTS_DIR",
    "TIMEZONE",
    "MockApiClient",
]



"""Defines the Auth class."""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

import aiohttp
from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning

from oocone import errors

if TYPE_CHECKING:
    from typing import Any

ROUTE_LOGIN = "/signinForm.php?mode=ok"
BEAUTIFULSOUP_PARSER = "html.parser"

logger = logging.getLogger(__name__)


class Auth:
    """Acquires authentication for the dashboard and makes authenticated requests."""

    def __init__(
        self,
        *,
        base_url: str,
        username: str,
        password: str,
        websession: aiohttp.ClientSession | None = None,
    ) -> None:
        """
        Create a new instance.

        Params:
            base_url: The URL on which the enocoo dashboard is available, e.g.
                ``https://ems123.enocoo.com:12345``. This will be used to initialize the
                [`base_url`][oocone.auth.Auth.base_url] attribute.
            username: The username to use for logging in to the dashboard.
            password: The password to use for logging in to the dashboard.
            websession: Allows customization of the used web session. If unset, a new session will
                be created internally.
        """
        self._base_url = base_url.rstrip("/")
        self.__username = username
        self.__password = password
        self._session = websession or aiohttp.ClientSession()
        self._logged_in = False

    @property
    def base_url(self) -> str:
        """The URL on which the enocoo dashboard is available."""
        return self._base_url

    @staticmethod
    def _response_indicates_not_logged_in(response: BeautifulSoup) -> bool:
        if response.title is not None and "abgemeldet" in response.title.text.lower():
            return True
        if response.find("input", {"type": "password"}) is not None:  # noqa: SIM103
            return True

        return False

    async def _login(self) -> None:
        async with self._session.post(
            self._base_url + ROUTE_LOGIN,
            data={"user": self.__username, "passwort": self.__password},
        ) as response:
            response_text = await response.text()

            soup = BeautifulSoup(response_text, features=BEAUTIFULSOUP_PARSER)
            if self._response_indicates_not_logged_in(soup):
                raise errors.AuthenticationFailed

    async def request(
        self, method: str, path: str, *, retry_with_login: bool = True, **kwargs: Any
    ) -> tuple[aiohttp.ClientResponse, BeautifulSoup]:
        """
        Make a request.

        Params:
            method: The [HTTP request method](https://en.wikipedia.org/wiki/HTTP#Request_methods) to
                use, e.g. ``GET`` or ``POST``.
            path: The path of the resource to request, relative to the
                [`base_ref`][oocone.Auth.__init__].
            retry_with_login: If this option is set and the response indicates that the resource
                could not be fetched due to not being logged in, the method will log in and retry
                the request.
            kwargs: Will be passed to [`ClientSession.request()`][aiohttp.ClientSession.request].

        Returns:
            the raw response
            allows scraping of the response

        """
        try:
            logger.debug("=>       %s %s/%s", method.upper(), self._base_url, path)
            response = await self._session.request(
                method,
                f"{self._base_url}/{path}",
                **kwargs,
            )
        except aiohttp.client_exceptions.ClientError as e:
            raise errors.ConnectionIssue from e

        logger.debug("<= (%s) %s %s/%s", response.status, method.upper(), self._base_url, path)
        soup = BeautifulSoup(await response.text(), BEAUTIFULSOUP_PARSER)

        if self._response_indicates_not_logged_in(soup):
            if retry_with_login:
                await self._login()
                return await self.request(method, path, retry_with_login=False, **kwargs)

            raise errors.AuthenticationFailed

        return (response, soup)


# Some of the responses returned by enocoo trigger a BeautifulSoup warning. So we disable the
# warning, at least for the current module.
warnings.filterwarnings(
    action="ignore",
    category=MarkupResemblesLocatorWarning,
    module=__name__,
    append=True,
)



"""Common fixtures and settings for all tests."""

import datetime as dt
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pytest
from aiohttp import web
from aiohttp.typedefs import Handler
from pytest_aiohttp.plugin import AiohttpClient

from . import TESTS_DIR, MockApiClient

pytest_plugins = ("pytest_asyncio",)

signed_in = web.AppKey("signed_in", bool)


@dataclass
class MockApiParams:
    """Parameters indicating how the Mock API works."""

    response_data_root_path: Path = TESTS_DIR / "data" / "responses"
    """The directory containing all response data sets."""

    dataset_name: str = "one_area"
    """Name of the dataset in use."""

    @property
    def response_data_path(self) -> Path:
        """Path to the response data set currently in use."""
        return self.response_data_root_path / self.dataset_name


MOCK_API_DATASET_NAMES = [child.name for child in MockApiParams().response_data_root_path.iterdir()]


@pytest.fixture
def mock_api_params() -> MockApiParams:
    """Return the MockApiParams for the current test."""
    return MockApiParams()


def _signin_form(mock_api_params: MockApiParams) -> Handler:
    async def handler(request: web.Request) -> web.Response:
        response_path = mock_api_params.response_data_path / "signinForm.failure.php"
        response = web.Response()

        if request.method == "POST":
            data = await request.post()
            username = data.get("user")
            password = data.get("passwort")
            credentials_correct = username == "correct" and password == "correct"  # noqa: S105

            if credentials_correct and request.query.get("mode") == "ok":
                response_path = mock_api_params.response_data_path / "signinForm.success.php"
                response.set_cookie("logged_in", "true")

        with Path.open(response_path, "rb") as f:
            response.body = f.read()

        return response

    return handler


def _response_from_file(
    response_path: Path | str, *, needs_login: bool = True, params: MockApiParams
) -> Handler:
    original_response_path = response_path

    async def handler(request: web.Request) -> web.Response:
        if needs_login and request.cookies.get("logged_in") != "true":
            response_path = params.response_data_path / "newMeterTable.notLoggedIn.php"
        else:
            response_path = params.response_data_path / original_response_path

        response = web.Response()
        with Path.open(response_path, "rb") as f:
            response.body = f.read()

        return response

    return handler


def _get_meter_data_with_param(mock_api_params: MockApiParams) -> Handler:
    async def handler(request: web.Request) -> web.Response:
        if request.cookies.get("logged_in") != "true":
            response_path = mock_api_params.response_data_path / "newMeterTable.notLoggedIn.php"
        else:
            q = request.query
            response_path = (
                mock_api_params.response_data_path
                / f"getMeterDataWithParam.{q['mClass']}.{q['AreaId']}.{q['from']}.{q['intVal']}.php"
            )

        response = web.Response()
        with Path.open(response_path, "rb") as f:
            response.body = f.read()

        return response

    return handler


def _get_pv_data_details(mock_api_params: MockApiParams) -> Handler:
    async def handler(request: web.Request) -> web.Response:
        if request.cookies.get("logged_in") != "true":
            response_path = mock_api_params.response_data_path / "newMeterTable.notLoggedIn.php"
        else:
            q = request.query
            response_path = (
                mock_api_params.response_data_path
                / f"getPVDataDetails.GesamtverbrauchUndErzeugung.{q['from']}.{q['intVal']}.php"
            )

        response = web.Response()
        with Path.open(response_path, "rb") as f:
            response.body = f.read()

        return response

    return handler


def _post_new_meter_table(mock_api_params: MockApiParams) -> Handler:
    async def handler(request: web.Request) -> web.Response:
        response = web.Response()
        response.body = _new_meter_table_body(
            logged_in=request.cookies.get("logged_in") == "true",
            date=dt.date.fromisoformat(str((await request.post())["dateParam"])),
            mock_api_params=mock_api_params,
        )
        return response

    return handler


def _new_meter_table_body(
    *, mock_api_params: MockApiParams, logged_in: bool, date: dt.date, time: dt.time | None = None
) -> bytes:
    if not logged_in:
        response_path = mock_api_params.response_data_path / "newMeterTable.notLoggedIn.php"
    else:
        response_path = mock_api_params.response_data_path / "newMeterTable.php"

    with Path.open(response_path, "rb") as f:
        body = f.read()

    body = body.replace(b"01.01.2021", f"{date.day:02}.{date.month:02}.{date.year:04}".encode())
    if time is not None:
        body = body.replace(b"12:34:56", time.isoformat("seconds").encode())

    return body


@pytest.fixture
async def mock_api(mock_api_params: MockApiParams, aiohttp_client: AiohttpClient) -> MockApiClient:
    """Return a mock API instance."""
    app = web.Application()
    app[signed_in] = False

    app.router.add_post("/signinForm.php", _signin_form(mock_api_params))
    app.router.add_get(
        "/php/getTrafficLightStatus.php",
        _response_from_file("getTrafficLightStatus.php", params=mock_api_params, needs_login=False),
    )
    app.router.add_get(
        "/php/getMeterDataWithParam.php", _get_meter_data_with_param(mock_api_params)
    )
    app.router.add_get("/php/getPVDataDetails.php", _get_pv_data_details(mock_api_params))
    app.router.add_post("/php/newMeterTable.php", _post_new_meter_table(mock_api_params))
    app.router.add_get(
        "/php/ownConsumption.php", _response_from_file("ownConsumption.php", params=mock_api_params)
    )
    return await aiohttp_client(app)


@pytest.fixture
async def mock_auth(mock_api: MockApiClient):  # type: ignore[no-untyped-def] # noqa: ANN201
    """Return an Auth instance accessing a mock API."""
    # Importing oocone directly inside conftest.py breaks the typeguard plugin for pytest,
    # so we import it lazily. Because of this, we cannot give a return type hint :(
    from oocone import Auth  # noqa: PLC0415

    return Auth(
        websession=mock_api.session,
        base_url=str(mock_api.server.make_url("/")),
        username="correct",
        password="correct",  # noqa: S106
    )


def pytest_collection_modifyitems(items: list[Any]) -> None:
    """Apply additional metadata to tests."""
    for item in items:
        if "mock_api" in item.fixturenames:
            item.add_marker(pytest.mark.mock_api)



"""Tests for the oocone.Enocoo.get_meter_table() method."""

import datetime as dt
from typing import Any

import pytest

from oocone import Auth, Enocoo
from oocone.model import MeterStatus, Quantity
from tests import TIMEZONE, conftest
from tests.conftest import MockApiParams

TODAY = dt.datetime.now(tz=TIMEZONE).date()
MIDNIGHT = dt.time(0, 0, 0)


def _expected_meter_table(date: dt.date, time: dt.time | None = None) -> list[MeterStatus]:
    if time is None:
        time = dt.time(12, 34, 56)
    expected_timestamp = dt.datetime.combine(date, time, tzinfo=TIMEZONE)
    return [
        MeterStatus(
            name="Verbrauch Kaltwasser H12W34 Bad",
            area="H12W34",
            meter_id="00000001",
            timestamp=expected_timestamp,
            reading=Quantity(value=1234.56, unit="m³"),
        ),
        MeterStatus(
            name="Verbrauch Kaltwasser H12W34 WC",
            area="H12W34",
            meter_id="00000002",
            timestamp=expected_timestamp,
            reading=Quantity(value=1234.56, unit="m³"),
        ),
        MeterStatus(
            name="Verbrauch Strom H12W34",
            area="H12W34",
            meter_id="00000003",
            timestamp=expected_timestamp,
            reading=Quantity(1234.56, unit="kWh"),
        ),
        MeterStatus(
            name="Verbrauch Wärme H12W34",
            area="H12W34",
            meter_id="00000004",
            timestamp=expected_timestamp,
            reading=Quantity(1234.56, unit="kWh"),
        ),
        MeterStatus(
            name="Verbrauch Warmwasser H12W34 Bad",
            area="H12W34",
            meter_id="00000005",
            timestamp=expected_timestamp,
            reading=Quantity(1234.56, unit="m³"),
        ),
        MeterStatus(
            name="Verbrauch Warmwasser H12W34 WC",
            area="H12W34",
            meter_id="00000006",
            timestamp=expected_timestamp,
            reading=Quantity(1234.56, unit="m³"),
        ),
    ]


async def test_without_params(mock_auth: Auth) -> None:
    """Check that the function returns data for the current day when called without parameters."""
    expected = _expected_meter_table(TODAY)

    enocoo = Enocoo(mock_auth, timezone=TIMEZONE)
    result = await enocoo.get_meter_table()

    assert result == expected


async def test_with_date_param(mock_auth: Auth) -> None:
    """Check that when called with a date param, the function returns data for that day."""
    date = dt.date(2012, 12, 12)
    expected = _expected_meter_table(date)

    enocoo = Enocoo(mock_auth, timezone=TIMEZONE)
    result = await enocoo.get_meter_table(date=date)

    assert result == expected


@pytest.mark.parametrize(
    ("allow_previous_day_until", "last_timestamp", "expected_result"),
    [
        # Shortly after midnight, the only data available are the ones for midnight. The enocoo
        # website does not list them on the page for the new day though, but on the page for the
        # previous day. The timestamp will still indicate the current date at 00:00.
        pytest.param(
            None,
            dt.datetime.combine(TODAY, MIDNIGHT),
            _expected_meter_table(TODAY, MIDNIGHT),
            id="midnight",
        ),
        # ------------------------------------------------------------------------------------------
        # ...except that is not entirely true: For the first 15 minutes after midnight, even the
        # data from midnight are not available, and instead we will see data from 15 minutes
        # earlier. Since that is not data from the current day, they are not returned by default.
        pytest.param(
            None,
            dt.datetime.combine(TODAY - dt.timedelta(days=1), dt.time(23, 45)),
            [],
            id="quarter-before-midnight-forbidden",
        ),
        # ------------------------------------------------------------------------------------------
        # ...except if the user explicitly opts into that.
        pytest.param(
            dt.time(23, 45),
            dt.datetime.combine(TODAY - dt.timedelta(days=1), dt.time(23, 45)),
            _expected_meter_table(TODAY - dt.timedelta(days=1), dt.time(23, 45)),
            id="quarter-before-midnight-allowed",
        ),
        # ------------------------------------------------------------------------------------------
        # ... but even if allow_previous_day_until is set: if the latest data are older than
        # allowed, they must not be returned.
        pytest.param(
            dt.time(23, 45),
            dt.datetime.combine(TODAY - dt.timedelta(days=1), dt.time(12, 00)),
            [],
            id="previous-noon-forbidden",
        ),
    ],
)
async def test_fallback_to_previous_day(
    allow_previous_day_until: dt.time | None,
    last_timestamp: dt.datetime,
    expected_result: list[MeterStatus],
    mock_auth: Auth,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Check that the method correctly implements fallback to the previous day."""
    original_new_meter_table_body = conftest._new_meter_table_body

    def patched_new_meter_table_body(
        *, logged_in: bool, date: dt.date, mock_api_params: MockApiParams
    ) -> bytes:
        if not logged_in:
            return original_new_meter_table_body(
                logged_in=logged_in, date=date, mock_api_params=mock_api_params
            )

        if date == TODAY:
            with (
                mock_api_params.response_data_path / "newMeterTable.noDataYetForCurrentDay.php"
            ).open("rb") as f:
                return f.read()
        elif date == TODAY - dt.timedelta(days=1):
            return original_new_meter_table_body(
                logged_in=logged_in,
                date=last_timestamp.date(),
                time=last_timestamp.time(),
                mock_api_params=mock_api_params,
            )
        else:
            msg = f"Date {date} should not be queried in this test"
            raise AssertionError(msg)

    monkeypatch.setattr(conftest, "_new_meter_table_body", patched_new_meter_table_body)

    enocoo = Enocoo(mock_auth, timezone=TIMEZONE)
    result = await enocoo.get_meter_table(allow_previous_day_until=allow_previous_day_until)

    assert result == expected_result


async def test_fallback_failing(
    mock_auth: Auth, mock_api_params: MockApiParams, monkeypatch: pytest.MonkeyPatch
) -> None:
    """
    Check that the method does not return data from a previous day.

    The fallback to the previous day is explained in ``test_fallback_to_midnight``. If the page for
    the previous day does not in fact return data for midnight of the current day, but in fact for
    the previous day, the method shall not return them.
    """
    original_new_meter_table_body = conftest._new_meter_table_body

    def patched_new_meter_table_body(*, logged_in: bool, date: dt.date, **kwargs: Any) -> bytes:
        if not logged_in:
            return original_new_meter_table_body(logged_in=logged_in, date=date, **kwargs)

        if date == TODAY:
            with (
                mock_api_params.response_data_path / "newMeterTable.noDataForCurrentDay.php"
            ).open("rb") as f:
                return f.read()
        elif date == TODAY - dt.timedelta(days=1):
            return original_new_meter_table_body(
                logged_in=logged_in, date=date, time=dt.time(10, 0, 0), **kwargs
            )
        else:
            msg = f"Date {date} should not be queried in this test"
            raise AssertionError(msg)

    monkeypatch.setattr(conftest, "_new_meter_table_body", patched_new_meter_table_body)

    enocoo = Enocoo(mock_auth, timezone=TIMEZONE)
    result = await enocoo.get_meter_table()

    assert result == []



import datetime as dt
import json
import logging
import re
from typing import cast

from bs4 import BeautifulSoup
from bs4 import Tag as HtmlTag

from oocone._internal.scrape_timeseries import (
    day_string_to_date,
    discard_unordered_hours,
    get_periods_per_hour,
    length_of_day,
)
from oocone.auth import Auth
from oocone.errors import UnexpectedResponse
from oocone.model import Area, Consumption, ConsumptionType

logger = logging.getLogger(__name__)

_CONSUMPTION_CLASSES = {
    ConsumptionType.ELECTRICITY: "Stromverbrauch",
    ConsumptionType.HEAT: "Waerme",
    ConsumptionType.WATER_COLD: "Kaltwasser",
    ConsumptionType.WATER_HOT: "Warmwasser",
}

_CONSUMPTION_UNITS = {
    ConsumptionType.ELECTRICITY: "kWh",
    ConsumptionType.HEAT: "kWh",
    ConsumptionType.WATER_COLD: "m³",
    ConsumptionType.WATER_HOT: "m³",
}

NUM_MONTHS = 12


async def get_areas(auth: Auth) -> list[Area]:
    logger.debug("Scraping areas from consumption page...")
    _, soup = await auth.request("GET", "php/ownConsumption.php")
    return _parse_areas(soup)


def _parse_areas(soup: BeautifulSoup) -> list[Area]:
    areas = []
    area_selectors = soup.select('label[for="residence"]')
    if area_selectors:
        logger.debug(
            "Found area selection dropdown on consumption page. Parsing the dropdown menu..."
        )
        for selector in area_selectors:
            areas += _parse_areas_for_selector(cast("HtmlTag", selector.parent))
    else:
        logger.debug(
            "Did not find area selection dropdown. Trying to find a reference to a single area..."
        )
        areas.append(_parse_single_area(soup))

    return areas


def _parse_areas_for_selector(selection_form: HtmlTag) -> list[Area]:
    areas = []
    for list_item in selection_form.find_all("li"):
        area_link = cast("HtmlTag", list_item).find("a")
        if not isinstance(area_link, HtmlTag):
            msg = f"Expected <a> Tag in area selector dropdown, got {type(area_link)} instead."
            raise UnexpectedResponse(msg)

        areas.append(
            Area(
                name=area_link.text.strip(),
                id=str(area_link.attrs["id"]),
                data_available_since=_parse_german_date(str(area_link.attrs["data-valfrom"])),
                data_available_until=_parse_german_date(str(area_link.attrs["data-valto"])),
            )
        )

    return areas


def _parse_single_area(soup: BeautifulSoup) -> Area:
    if (
        match1 := re.search(
            r'var chosenResidenceId = "(\d+)";',
            "\n\n".join(script.text for script in soup.find_all("script")),
        )
    ) and (
        match2 := re.search(
            r"Wohneinheit: (H\d+W\d+)"
            r" \(Zugreifbarer Zeitraum: (\d{2}\.\d{2}\.\d{4}) - (\d{2}\.\d{2}\.\d{4})",
            soup.text,
        )
    ):
        return Area(
            id=match1.group(1),
            name=match2.group(1),
            data_available_since=_parse_german_date(match2.group(2)),
            data_available_until=_parse_german_date(match2.group(3)),
        )

    msg = "Failed to scrape single area ID from page"
    raise UnexpectedResponse(msg)


def _parse_german_date(date_string: str) -> dt.date:
    return dt.datetime.strptime(date_string, "%d.%m.%Y").date()  # noqa: DTZ007


async def get_daily_consumption(
    consumption_type: ConsumptionType, area_id: str, date: dt.date, timezone: dt.tzinfo, auth: Auth
) -> list[Consumption]:
    logger.debug(
        "Scraping daily %s consumption on %s for area with ID %s...",
        consumption_type,
        date,
        area_id,
    )
    response, _ = await auth.request(
        "GET",
        "php/getMeterDataWithParam.php",
        params={
            "AreaId": area_id,
            "from": date.isoformat(),
            "intVal": "Tag",
            "mClass": _CONSUMPTION_CLASSES[consumption_type],
        },
    )
    return _parse_daily_consumption(
        daily_consumption_json=await response.text(),
        unit=_CONSUMPTION_UNITS[consumption_type],
        date=date,
        timezone=timezone,
        values_are_integrated=(consumption_type == ConsumptionType.HEAT),
    )


async def get_monthly_consumption(
    consumption_type: ConsumptionType, area_id: str, date: dt.date, timezone: dt.tzinfo, auth: Auth
) -> list[Consumption]:
    logger.debug(
        "Scraping monthly %s consumption on %s for area with ID %s...",
        consumption_type,
        date,
        area_id,
    )
    response, _ = await auth.request(
        "GET",
        "php/getMeterDataWithParam.php",
        params={
            "AreaId": area_id,
            "from": date.isoformat(),
            "intVal": "Monat",
            "mClass": _CONSUMPTION_CLASSES[consumption_type],
        },
    )
    return _parse_monthly_consumption(
        monthly_consumption_json=await response.text(),
        unit=_CONSUMPTION_UNITS[consumption_type],
        year=date.year,
        month=date.month,
        timezone=timezone,
    )


async def get_yearly_consumption(
    consumption_type: ConsumptionType,
    area_id: str,
    year_number: int,
    timezone: dt.tzinfo,
    auth: Auth,
) -> list[Consumption]:
    logger.debug(
        "Scraping yearly %s consumption in %s for area with ID %s...",
        consumption_type,
        year_number,
        area_id,
    )
    response, _ = await auth.request(
        "GET",
        "php/getMeterDataWithParam.php",
        params={
            "AreaId": area_id,
            "from": f"{year_number}-01-01",
            "intVal": "Jahr",
            "mClass": _CONSUMPTION_CLASSES[consumption_type],
        },
    )
    return _parse_yearly_consumption(
        yearly_consumption_json=await response.text(),
        unit=_CONSUMPTION_UNITS[consumption_type],
        year_number=year_number,
        timezone=timezone,
    )


def _parse_daily_consumption(
    *,
    daily_consumption_json: str,
    values_are_integrated: bool,
    unit: str,
    date: dt.date,
    timezone: dt.tzinfo,
) -> list[Consumption]:
    results = []

    json_data = json.loads(daily_consumption_json)
    hours = json_data[1]
    values = json_data[0]

    last_time = dt.datetime(
        date.year, date.month, date.day, hour=0, minute=0, second=0, tzinfo=timezone
    )
    last_hour = None
    last_value = 0

    hours, values = discard_unordered_hours(hours, values, date, description="consumption")
    periods_per_hour = get_periods_per_hour(hours)

    for hour, value in zip(hours, values, strict=False):
        period = periods_per_hour[hour]

        if hour == last_hour:
            time = last_time + period
        else:
            time = last_time.replace(hour=hour, minute=0)

        if values_are_integrated:
            consumption_value = float(value) - float(last_value)
        else:
            consumption_value = float(value)

        consumption = Consumption(start=time, period=period, value=consumption_value, unit=unit)
        results.append(consumption)

        last_time = time
        last_hour = hour
        last_value = value

    return results


def _parse_monthly_consumption(
    *,
    monthly_consumption_json: str,
    unit: str,
    year: int,
    month: int,
    timezone: dt.tzinfo,
) -> list[Consumption]:
    results = []

    json_data = json.loads(monthly_consumption_json)
    values = json_data[0]
    dates = [day_string_to_date(day_string, month, year) for day_string in json_data[1]]

    for date, value in zip(dates, values, strict=False):
        consumption = Consumption(
            start=dt.datetime.combine(date, dt.time(0, 0), tzinfo=timezone),
            period=length_of_day(date, timezone),
            value=value,
            unit=unit,
        )
        results.append(consumption)

    return results


def _parse_yearly_consumption(
    *,
    yearly_consumption_json: str,
    unit: str,
    year_number: int,
    timezone: dt.tzinfo,
) -> list[Consumption]:
    results = []

    month_names_to_number = {
        "Jan.": 1,
        "Feb.": 2,
        "Mär.": 3,
        "Apr.": 4,
        "Mai": 5,
        "Jun.": 6,
        "Jul.": 7,
        "Aug.": 8,
        "Sep.": 9,
        "Okt.": 10,
        "Nov.": 11,
        "Dez.": 12,
    }

    json_data = json.loads(yearly_consumption_json)
    month_names = json_data[1]
    values = json_data[0]

    for month_name, value in zip(month_names, values, strict=False):
        month = month_names_to_number[month_name]

        begin_of_month = dt.date(year_number, month, 1)
        begin_of_next_month = dt.date(
            year_number if month < NUM_MONTHS else year_number + 1,
            month + 1 if month < NUM_MONTHS else 1,
            1,
        )
        consumption = Consumption(
            start=dt.datetime.combine(begin_of_month, dt.time(tzinfo=timezone)),
            period=begin_of_next_month - begin_of_month,
            value=value,
            unit=unit,
        )
        results.append(consumption)

    return results



import datetime as dt
import json
import logging
from typing import Any

from oocone._internal.scrape_timeseries import (
    day_string_to_date,
    discard_unordered_hours,
    get_periods_per_hour,
    length_of_day,
)
from oocone.auth import Auth
from oocone.model import PhotovoltaicSummary, Quantity

logger = logging.getLogger(__name__)


async def get_daily_photovoltaic_data(
    date: dt.date, timezone: dt.tzinfo, auth: Auth
) -> list[PhotovoltaicSummary]:
    logger.debug("Scraping daily photovoltaic data on %s...", date)
    response, _ = await auth.request(
        "GET",
        "php/getPVDataDetails.php",
        params={
            "from": date.isoformat(),
            "intVal": "Tag",
            "diagType": "GesamtverbrauchUndErzeugung",
        },
    )
    return _parse_daily_photovoltaic_data(
        data=await response.text(),
        unit="kWh",
        date=date,
        timezone=timezone,
    )


async def get_monthly_photovoltaic_data(
    date: dt.date, timezone: dt.tzinfo, auth: Auth
) -> list[PhotovoltaicSummary]:
    logger.debug("Scraping monthly photovoltaic data on %s...", date)
    response, _ = await auth.request(
        "GET",
        "php/getPVDataDetails.php",
        params={
            "from": date.isoformat(),
            "intVal": "Monat",
            "diagType": "GesamtverbrauchUndErzeugung",
        },
    )
    return _parse_monthly_photovoltaic_data(
        data=await response.text(),
        unit="kWh",
        year=date.year,
        month=date.month,
        timezone=timezone,
    )


def _parse_daily_photovoltaic_data(
    *,
    data: str,
    unit: str,
    date: dt.date,
    timezone: dt.tzinfo,
) -> list[PhotovoltaicSummary]:
    results = []

    json_data = json.loads(data)
    raw_values = zip(*(json_data[0:4]), strict=True)
    hours = json_data[4]

    last_time = dt.datetime(
        date.year, date.month, date.day, hour=0, minute=0, second=0, tzinfo=timezone
    )
    last_hour = None

    hours, values = discard_unordered_hours(
        hours, raw_values, date, description="photovoltaic data"
    )
    periods_per_hour = get_periods_per_hour(hours)

    for hour, value in zip(hours, values, strict=False):
        period = periods_per_hour[hour]

        if hour == last_hour:
            time = last_time + period
        else:
            time = last_time.replace(hour=hour, minute=0)

        summary = PhotovoltaicSummary(
            start=time,
            period=period,
            consumption=Quantity(value=float(value[0]), unit=unit),
            generation=Quantity(value=float(value[1]), unit=unit),
            self_sufficiency=_quantity_or_none(value[2], unit="%"),
            own_consumption=_quantity_or_none(value[3], unit="%"),
        )
        results.append(summary)

        last_time = time
        last_hour = hour

    return results


def _parse_monthly_photovoltaic_data(
    *,
    data: str,
    unit: str,
    year: int,
    month: int,
    timezone: dt.tzinfo,
) -> list[PhotovoltaicSummary]:
    results = []

    json_data = json.loads(data)
    values = zip(*(json_data[0:4]), strict=True)
    dates = [day_string_to_date(day_string, month, year) for day_string in json_data[4]]

    for date, value in zip(dates, values, strict=False):
        summary = PhotovoltaicSummary(
            start=dt.datetime.combine(date, dt.time(0, 0), tzinfo=timezone),
            period=length_of_day(date, timezone=timezone),
            consumption=Quantity(value=float(value[0]), unit=unit),
            generation=Quantity(value=float(value[1]), unit=unit),
            self_sufficiency=_quantity_or_none(value[2], unit="%"),
            own_consumption=_quantity_or_none(value[3], unit="%"),
        )
        results.append(summary)

    return results


def _quantity_or_none(value: Any, unit: str) -> Quantity | None:
    if value is None:
        return None
    return Quantity(value=float(value), unit=unit)



"""Tests for oocone.Auth."""

import pytest

from oocone import Auth, errors
from tests import MockApiClient


async def test_login_successful(mock_api: MockApiClient) -> None:
    """Check that Auth._login() does not throw an exception with correct credentials."""
    auth = Auth(
        websession=mock_api.session,
        base_url=str(mock_api.server.make_url("/")),
        username="correct",
        password="correct",  # noqa: S106
    )
    await auth._login()


async def test_login_raise_on_failure(mock_api: MockApiClient) -> None:
    """Check that Auth._login() raises an exception for incorrect credentials."""
    auth = Auth(
        websession=mock_api.session,
        base_url=str(mock_api.server.make_url("/")),
        username="correct",
        password="incorrect",  # noqa: S106
    )
    with pytest.raises(errors.AuthenticationFailed):
        await auth._login()



"""Tests for oocone._internal.scrape_timeseries."""

import datetime as dt
from itertools import chain, repeat

from oocone._internal import scrape_timeseries


def test_get_periods_per_hour_four_points_per_hour() -> None:
    """Check that with four points per hours, for 24 hours straight, all periods are 15 minutes."""
    hours = list(chain(*(repeat(hour, 4) for hour in range(24))))
    expected_periods = {hour: dt.timedelta(minutes=15) for hour in range(24)}

    periods = scrape_timeseries.get_periods_per_hour(hours)
    assert periods == expected_periods


def test_get_periods_per_hour_more_points_per_hour() -> None:
    """Check that if an hour contains five or six data points, periods are 12 and minutes."""
    hours = [0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3]
    expected_periods = {
        0: dt.timedelta(minutes=15),
        1: dt.timedelta(minutes=12),
        2: dt.timedelta(minutes=10),
        3: dt.timedelta(minutes=15),
    }

    periods = scrape_timeseries.get_periods_per_hour(hours)
    assert periods == expected_periods


def test_get_periods_per_hour_less_points_per_hour() -> None:
    """Check that if an hour contains less than four data points, periods are still 15 minutes."""
    hours = [0, 0, 0, 0, 1, 1, 2, 2, 2]
    expected_periods = {
        0: dt.timedelta(minutes=15),
        1: dt.timedelta(minutes=15),
        2: dt.timedelta(minutes=15),
    }

    periods = scrape_timeseries.get_periods_per_hour(hours)
    assert periods == expected_periods



"""Tests for the oocone.Enocoo.get_traffic_light_status() method."""

from oocone import Auth, Enocoo
from oocone.model import Quantity, TrafficLightColor
from tests import TIMEZONE


async def test_happy_path(mock_auth: Auth) -> None:
    """Check that Enocoo.get_traffic_light_status successfully returns for mock API data."""
    enocoo = Enocoo(mock_auth, TIMEZONE)
    result = await enocoo.get_traffic_light_status()
    assert isinstance(result.color, TrafficLightColor)
    assert isinstance(result.current_energy_price, Quantity)
    assert isinstance(result.current_energy_price.value, float)



"""Defines exceptions specific to this module."""


class OoconeError(Exception):
    """Base class for oocone."""


class ConnectionIssue(OoconeError):
    """Could not connect to the enocoo website."""


class AuthenticationFailed(OoconeError):
    """Authentication has failed."""


class UnexpectedResponse(OoconeError):
    """Got unexpected data back from the enocoo website."""


class OoconeMisuse(OoconeError):
    """The oocone API was misused."""



"""Contains various functions for handling quirks of the enocoo portal."""

import asyncio
import dataclasses
import datetime as dt
import itertools
from collections.abc import Awaitable, Callable

from oocone.model import Consumption


async def get_off_by_one_compensated_data(
    fetch_data: Callable[[dt.date], Awaitable[list[Consumption]]],
    date: dt.date,
    timezone: dt.tzinfo,
) -> list[Consumption]:
    """
    Compensates for an off by one data point bug in the enocoo EMS.

    The enocoo EMS has this bug where energy usage is off by one sample, i.e.  15 minutes. Because
    of this especially, the sum of the daily consumption numbers does does not add up to the actual
    difference of the meter at 00:00 on two consecutive days. Concretely this means that the first
    consumption sample of a given day belongs to the previous day (and in turn that any consumption
    after 23:45 of any given day shows up in the statistics for the day *after* that).
    Unfortunately, this means that for a complete day, we will need to make two requests towards the
    enocoo API instead of one.
    To go as easy as possible on the API, two measures are implemented:

     2. checks against the current date: when requesting data, we never ask for data of a day that
        has not begun yet. This means that we only ask query data for "the next day" when filling
        in the last hour of the day, not any time before that.

    In addition, oocone users should implement a caching mechanism, as `ha_enocoo` does for example.
    """
    original_day = date
    next_day = original_day + dt.timedelta(days=1)
    now = dt.datetime.now(tz=timezone)

    async def fetch_data_omitting_future(date: dt.date) -> list[Consumption]:
        if date > now.date():  # date is in the future, no point in querying enocoo for data
            return []

        return await fetch_data(date)

    off_by_one_data, next_day_data = await asyncio.gather(
        fetch_data_omitting_future(original_day), fetch_data_omitting_future(next_day)
    )
    off_by_one_data += next_day_data[:1]

    return [
        dataclasses.replace(
            point_with_correct_data,
            start=point_with_correct_timing.start,
            period=point_with_correct_timing.period,
        )
        for point_with_correct_timing, point_with_correct_data in itertools.pairwise(
            off_by_one_data
        )
    ]



"""Utilities for scraping HTML tables."""

from dataclasses import dataclass
from typing import cast

from bs4 import Tag as HtmlTag

from oocone import errors

DataRow = dict[str, str]


@dataclass
class Table:
    columns: list[str]
    rows: list[DataRow]


def parse_table(html_table: HtmlTag) -> Table:
    rows = cast("list[HtmlTag]", html_table.find_all("tr"))
    for row in rows:
        if not isinstance(row, HtmlTag):
            msg = f"Expected tr as Tag but found {type(row)}"
            raise errors.UnexpectedResponse(msg)

    columns = _parse_header_row(rows[0])
    data_rows = [_parse_data_row(r, columns) for r in rows[1:]]

    return Table(columns=columns, rows=data_rows)


def _parse_header_row(row: HtmlTag) -> list[str]:
    return [cell.text for cell in row.find_all("th")]


def _parse_data_row(row: HtmlTag, columns: list[str]) -> DataRow:
    row_cells = row.find_all("td")
    return {header: cell.text for header, cell in zip(columns, row_cells, strict=False)}



"""Tests for oocone._internal.html_table."""

from pathlib import Path

import pytest
from bs4 import BeautifulSoup, Tag

from oocone._internal.html_table import Table, parse_table
from oocone.enocoo import BEAUTIFULSOUP_PARSER
from tests.conftest import MockApiParams


def __testdata(data_id: str) -> tuple[str, Table]:
    if data_id == "synthetic":
        html = """\
            <table class="table table-striped">
                <tr>
                    <th>Header 1</th>
                    <th>Header 2</th>
                    <th>Header 3</th>
                </tr>
                <tr>
                    <td>Row 1, Data Point 1</td>
                    <td>Row 1, Data Point 2</td>
                    <td>Row 1, Data Point 3</td>
                </tr>
                <tr>
                    <td>Row 2, Data Point 1</td>
                    <td>Row 2, Data Point 2</td>
                    <td>Row 2, Data Point 3</td>
                </tr>
            </table>
            """
        expected_table = Table(
            columns=["Header 1", "Header 2", "Header 3"],
            rows=[
                {
                    "Header 1": "Row 1, Data Point 1",
                    "Header 2": "Row 1, Data Point 2",
                    "Header 3": "Row 1, Data Point 3",
                },
                {
                    "Header 1": "Row 2, Data Point 1",
                    "Header 2": "Row 2, Data Point 2",
                    "Header 3": "Row 2, Data Point 3",
                },
            ],
        )
    elif data_id == "newMeterTable.php":
        html = Path.open(
            MockApiParams().response_data_path / "newMeterTable.php", encoding="utf-8"
        ).read()
        expected_table = Table(
            columns=["Fläche", "Bezeichnung", "Zähler-Nr.", "Zeitpunkt", "Zählerstand", "Einheit"],
            rows=[
                {
                    "Fläche": "H12W34",
                    "Bezeichnung": "Verbrauch Kaltwasser H12W34 Bad",
                    "Zähler-Nr.": "00000001",
                    "Zeitpunkt": "01.01.2021 12:34:56",
                    "Zählerstand": "1.234,56",
                    "Einheit": "m3",
                },
                {
                    "Fläche": "H12W34",
                    "Bezeichnung": "Verbrauch Kaltwasser H12W34 WC",
                    "Zähler-Nr.": "00000002",
                    "Zeitpunkt": "01.01.2021 12:34:56",
                    "Zählerstand": "1.234,56",
                    "Einheit": "m3",
                },
                {
                    "Fläche": "H12W34",
                    "Bezeichnung": "Verbrauch Strom H12W34",
                    "Zähler-Nr.": "00000003",
                    "Zeitpunkt": "01.01.2021 12:34:56",
                    "Zählerstand": "1.234,56",
                    "Einheit": "kWh",
                },
                {
                    "Fläche": "H12W34",
                    "Bezeichnung": "Verbrauch Wärme H12W34",
                    "Zähler-Nr.": "00000004",
                    "Zeitpunkt": "01.01.2021 12:34:56",
                    "Zählerstand": "1.234,56",
                    "Einheit": "kWh",
                },
                {
                    "Fläche": "H12W34",
                    "Bezeichnung": "Verbrauch Warmwasser H12W34 Bad",
                    "Zähler-Nr.": "00000005",
                    "Zeitpunkt": "01.01.2021 12:34:56",
                    "Zählerstand": "1.234,56",
                    "Einheit": "m3",
                },
                {
                    "Fläche": "H12W34",
                    "Bezeichnung": "Verbrauch Warmwasser H12W34 WC",
                    "Zähler-Nr.": "00000006",
                    "Zeitpunkt": "01.01.2021 12:34:56",
                    "Zählerstand": "1.234,56",
                    "Einheit": "m3",
                },
            ],
        )
    else:
        msg = f'Unknown data ID "{data_id}"'
        raise ValueError(msg)

    return html, expected_table


@pytest.mark.parametrize(
    "data_id",
    [
        "synthetic",
        "newMeterTable.php",
    ],
)
def test_parse_table(data_id: str) -> None:
    """Check that a given HTML string is parsed as a specific expected Table instance."""
    html, expected_table = __testdata(data_id)
    html_table = BeautifulSoup(html, features=BEAUTIFULSOUP_PARSER).find("table")
    assert isinstance(html_table, Tag)
    parsed_table = parse_table(html_table)

    assert parsed_table == expected_table



import datetime as dt
import logging

from bs4 import Tag

from oocone import errors
from oocone._internal.html_table import parse_table
from oocone.auth import Auth
from oocone.model import MeterStatus, Quantity

logger = logging.getLogger(__name__)


async def get_meter_table(date: dt.date, timezone: dt.tzinfo, auth: Auth) -> list[MeterStatus]:
    logger.debug("Scraping meter table for %s...", date)
    _, soup = await auth.request(
        "POST",
        "php/newMeterTable.php",
        data={"dateParam": date.isoformat()},
    )
    html_table = soup.find("table")
    if not isinstance(html_table, Tag):
        msg = f"Expected table as Tag but found {type(html_table)}"
        raise errors.UnexpectedResponse(msg)
    meter_table = parse_table(html_table)

    result = []
    for row in meter_table.rows:
        try:
            if row["Zeitpunkt"] == "" or row["Einheit"] == "":
                continue

            meter_status = MeterStatus(
                name=row["Bezeichnung"],
                area=row["Fläche"],
                meter_id=row["Zähler-Nr."],
                timestamp=_parse_timestamp(row["Zeitpunkt"], timezone),
                reading=Quantity(
                    value=_parse_reading(row["Zählerstand"]),
                    unit=_parse_unit(row["Einheit"]),
                ),
            )
        except Exception as e:
            raise errors.UnexpectedResponse from e

        result.append(meter_status)

    return result


def _parse_timestamp(timestamp: str, timezone: dt.tzinfo) -> dt.datetime:
    dateformat = r"%d.%m.%Y %H:%M:%S"
    return dt.datetime.strptime(timestamp, dateformat).replace(tzinfo=timezone)


def _parse_reading(reading: str) -> float:
    # The reading uses german number formatting with a comma as the decimal separator and a
    # dot as the thousands separator. The convention used by float is a dot as the decimal
    # separator and comma as the thousands separator.abs

    reading = reading.replace(".", "")  # we don't need a thousands separator here
    reading = reading.replace(",", ".")

    return float(reading)


def _parse_unit(text: str) -> str:
    if text == "m3":
        result = "m³"
    else:
        result = text

    return result



"""Module containing the main API entry points for this module."""

from __future__ import annotations

import datetime as dt
import logging
import warnings
from typing import TYPE_CHECKING, Literal

from oocone import errors
from oocone._internal import (
    quirks,
    scrape_consumption,
    scrape_meter_table,
    scrape_photovoltaic,
    scrape_traffic_light,
)

if TYPE_CHECKING:
    from oocone.auth import Auth
    from oocone.model import (
        Area,
        Consumption,
        ConsumptionType,
        MeterStatus,
        PhotovoltaicSummary,
        TrafficLightStatus,
    )

ROUTE_LOGIN = "/signinForm.php?mode=ok"
BEAUTIFULSOUP_PARSER = "html.parser"

logger = logging.getLogger(__name__)


class Enocoo:
    """Provides access to the data accessible via the enocoo Web interface."""

    def __init__(self, auth: Auth, timezone: dt.tzinfo) -> None:
        """
        Initialize the API and store the auth so we can make requests.

        Params:
            auth: Indicates how to contact the enocoo dashboard, including URL and credentials.
            timezone: The timezone in which the building of the energy management system is located.
        """
        self.auth = auth
        self.timezone = timezone

    async def get_traffic_light_status(self) -> TrafficLightStatus:
        """
        Return the status of the energy traffic light.

        Returns:
            Current status of the energy traffic light.

        """
        return await scrape_traffic_light.get_traffic_light_status(self.auth)

    async def get_meter_table(
        self,
        date: dt.date | None = None,
        allow_previous_day_until: dt.time | None = None,
    ) -> list[MeterStatus]:
        """
        Return the status of all individual consumption meters available in the dashboard.

        Params:
            date: The date for which meter data should be fetched. If left to None: the current day.
            allow_previous_day_until: If set, data from the previous day might be returned if the
                timestamp is at least the value of this parameter.

        Returns:
            For all individual consumption meters available in the dashboard, their latest status
                on the given date.

        """
        if date is None:
            date = dt.datetime.now(tz=self.timezone).date()

        meter_table = await scrape_meter_table.get_meter_table(
            date=date,
            timezone=self.timezone,
            auth=self.auth,
        )

        if not meter_table:
            day_before = date - dt.timedelta(days=1)
            meter_table_day_before = await scrape_meter_table.get_meter_table(
                date=day_before,
                timezone=self.timezone,
                auth=self.auth,
            )

            def status_is_relevant(status: MeterStatus) -> bool:
                relevant = False
                if status.timestamp.date() == date:
                    # status is from today -> relevant
                    relevant = True
                elif (
                    allow_previous_day_until is not None
                    and status.timestamp.time() >= allow_previous_day_until
                    and status.timestamp.date() == (date - dt.timedelta(days=1))
                ):
                    # status is declared relevant explicitly by user wishes
                    relevant = True
                return relevant

            meter_table = [s for s in meter_table_day_before if status_is_relevant(s)]
            logger.debug(
                "Found %s relevant meter statuses from the day before (%s), returning those...",
                len(meter_table),
                day_before,
            )

        return meter_table

    async def get_areas(self) -> list[Area]:
        """
        Get all available areas via the dashboard.

        Returns:
            A list of all areas for which data can be read from the enocoo dashboard.

        """
        return await scrape_consumption.get_areas(auth=self.auth)

    async def get_individual_consumption(
        self,
        consumption_type: ConsumptionType,
        *,
        during: dt.date,
        interval: Literal["day", "month", "year"],
        area_id: str,
        compensate_off_by_one: bool | None = None,
    ) -> list[Consumption]:
        """
        Return individual consumption statistics for a given meter type.

        Params:
            consumption_type: The kind of resource consumption to fetch data for (e.g. electricity,
                heat etc.).
            during: The date for which to fetch consumption data.
            interval: The length of interval (around ``date``) for which consumption data shall be
                fetched.
            area_id: Individual consumption is always tied to a specific area (a list of which you
                can obtain via [`get_areas()`][oocone.enocoo.Enocoo.get_areas]). You need to give
                the ID of the area that data shall be fetched for here.
            compensate_off_by_one: In some cases, the dates for data returned by the enocoo
                dashboard are off by one time interval. oocone offers the possibility of correcting
                this client-side. By **not** setting this parameter, this compensation will be
                activated in all cases where off-by-one errors are known. You can turn this off by
                setting this parameter to `False`.

        Returns:
            A list of consumption data points per time interval.

        """
        if compensate_off_by_one is None:
            compensate_off_by_one = interval == "day"

        if compensate_off_by_one and interval != "day":
            warnings.warn(
                "Data points in the enocoo dashboard are off by one. Compensation for this is only"
                " available when fetching daily consumption statistics.",
                stacklevel=1,
            )
            compensate_off_by_one = False

        async def fetch_data(date: dt.date) -> list[Consumption]:
            return await self._get_individual_consumption_uncompensated(
                consumption_type=consumption_type, during=date, interval=interval, area_id=area_id
            )

        if compensate_off_by_one:
            result = await quirks.get_off_by_one_compensated_data(
                fetch_data=fetch_data,
                date=during,
                timezone=self.timezone,
            )
        else:
            result = await fetch_data(date=during)

        return result

    async def _get_individual_consumption_uncompensated(
        self,
        consumption_type: ConsumptionType,
        during: dt.date,
        interval: Literal["day", "month", "year"],
        area_id: str,
    ) -> list[Consumption]:
        match interval:
            case "day":
                return await scrape_consumption.get_daily_consumption(
                    consumption_type=consumption_type,
                    area_id=area_id,
                    date=during,
                    timezone=self.timezone,
                    auth=self.auth,
                )
            case "month":
                return await scrape_consumption.get_monthly_consumption(
                    consumption_type=consumption_type,
                    area_id=area_id,
                    date=during,
                    timezone=self.timezone,
                    auth=self.auth,
                )
            case "year":
                return await scrape_consumption.get_yearly_consumption(
                    consumption_type=consumption_type,
                    area_id=area_id,
                    year_number=during.year,
                    timezone=self.timezone,
                    auth=self.auth,
                )
            case _:
                msg = f'Illegal interval "{interval}"'
                raise errors.OoconeMisuse(msg)

    async def get_quarter_photovoltaic_data(
        self,
        during: dt.date,
        interval: Literal["day", "month"],
    ) -> list[PhotovoltaicSummary]:
        """
        Return photovoltaic data for the whole quarter.

        Params:
            during: The date for which to fetch consumption data.
            interval: The length of interval (around ``date``) for which consumption data shall be
                fetched.

        Returns:
            A list of (quarter) photovoltaic data points during the given interval.

        """
        match interval:
            case "day":
                return await scrape_photovoltaic.get_daily_photovoltaic_data(
                    date=during, timezone=self.timezone, auth=self.auth
                )
            case "month":
                return await scrape_photovoltaic.get_monthly_photovoltaic_data(
                    date=during, timezone=self.timezone, auth=self.auth
                )
            case _:
                msg = f'Illegal interval "{interval}"'
                raise errors.OoconeMisuse(msg)



import datetime as dt
import logging
from collections import Counter
from collections.abc import Iterable

logger = logging.getLogger(__name__)


def get_periods_per_hour(hours: list[int]) -> dict[int, dt.timedelta]:
    one_hour = dt.timedelta(hours=1)
    max_period = dt.timedelta(minutes=15)

    return {
        hour: min(one_hour / measurements_per_hour, max_period)
        for hour, measurements_per_hour in Counter(hours).items()
    }


def discard_unordered_hours[ValueT](
    hours: Iterable[int],
    values: Iterable[ValueT],
    date: dt.date,
    *,
    description: str,
    stop_after_midnight: bool = True,
) -> tuple[list[int], list[ValueT]]:
    filtered_hours = []
    filtered_values = []

    last_hour = None

    for hour, value in zip(hours, values, strict=False):
        if stop_after_midnight and last_hour == 23 and hour != last_hour:  # noqa: PLR2004
            # hour after midnight -> discard
            break
        if last_hour is not None and hour < last_hour:
            logger.warning(
                "Hours are unordered in daily %s reading for %s: hour %s follows hour %s."
                " %s metric is discarded.",
                description,
                date.isoformat(),
                hour,
                last_hour,
                description.capitalize(),
            )
            continue

        filtered_hours.append(hour)
        filtered_values.append(value)
        last_hour = hour

    return filtered_hours, filtered_values


def day_string_to_date(day_str: str, month: int, year: int) -> dt.date:
    r"""
    Convert a day string to an actual date.

    Args:
        day_str: a string returned by Enocoo APIs. Example: "So.\n 01."
        month: current month (1-2)
        year: current year

    """
    day = int(day_str.splitlines()[-1].strip().lstrip("0").rstrip("."))
    return dt.date(year, month, day)


def length_of_day(day: dt.date, timezone: dt.tzinfo) -> dt.timedelta:
    next_day = day + dt.timedelta(days=1)
    midnight = dt.datetime.combine(day, dt.time(0, 0), tzinfo=timezone)
    next_midnight = dt.datetime.combine(next_day, dt.time(0, 0), tzinfo=timezone)

    return next_midnight - midnight



import logging
from typing import Any

from oocone import errors
from oocone.auth import Auth
from oocone.model import UNKNOWN, Quantity, TrafficLightColor, TrafficLightStatus, UnknownT

logger = logging.getLogger(__name__)


async def get_traffic_light_status(auth: Auth) -> TrafficLightStatus:
    """Return the status of the energy traffic light."""
    logger.debug("Scraping current traffic light status...")
    response, _ = await auth.request("GET", "php/getTrafficLightStatus.php")

    try:
        # We parse the response as JSON, even though the Content-Type header might indicate
        # otherwise.
        response_data = await response.json(content_type=None)
    except Exception as e:
        raise errors.UnexpectedResponse from e

    return TrafficLightStatus(
        color=_parse_color(response_data),
        current_energy_price=_parse_current_energy_price(response_data),
    )


def _parse_color(response_data: dict) -> TrafficLightColor | UnknownT:
    try:
        raw = _extract_key_from_response(response_data, "color")
    except KeyError as e:
        logger.warning(e)
        return UNKNOWN

    if raw == "rot":
        return TrafficLightColor.RED
    if raw == "gelb":
        return TrafficLightColor.YELLOW
    if raw == r"grün":
        return TrafficLightColor.GREEN

    logger.warning('Got unexpected color: "%s", raw')
    return UNKNOWN


def _parse_current_energy_price(response_data: dict) -> Quantity | UnknownT:
    try:
        raw = _extract_key_from_response(response_data, "currentEnergyprice")
    except KeyError as e:
        logger.warning(e)
        return UNKNOWN

    try:
        result = Quantity(
            value=float(raw),
            unit="ct/kWh",  # Even though the API does not tell us that directly, the unit is Cents
        )
    except ValueError:
        logger.warning("Could not parse energy price %s as a number", raw)
        return UNKNOWN

    return result


def _extract_key_from_response(response_data: dict[str, Any], key: str) -> Any:
    try:
        result = response_data[key]
    except KeyError:
        msg = f'API response does not contain key "{key}".\nResponse data:\n{response_data}'
        raise KeyError(msg) from None

    return result
